#import "Tty.h"
#import <stdio.h>
#import <sys/file.h>
#import <string.h>
#import <libc.h>

@implementation Tty
  
+ newDevice:(char *)devname access:(int)flags
{
  int cc;

  self = [super new];
  fd = open(devname, flags, 0);
  if (fd<0) {
    fprintf(stderr,"Couldn't open %s\n",devname); perror("Unix error follows");
    [super free];
    return nil;
  }
  cc=ioctl(fd,TIOCGETP,&origState);
  currentState = origState;
  return self;
}


+ newReadDevice:(char *)devname
{
  return [self newDevice:devname access:O_RDONLY];
}

+ newWriteDevice:(char *)devname
{
  return [self newDevice:devname access:O_WRONLY];
}

+ newReadWriteDevice:(char *)devname
{
  return [self newDevice:devname access:O_RDWR];
}

+ newExistingfd:(int)thisfd  // for already existing file descriptor (eg. 0=stdin 1=stdout 2=stderr)
{
  int cc;

  self = [super new];
  fd = thisfd;
  cc=ioctl(fd,TIOCGETP,&origState);
  currentState = origState;
  return self;
}

- reset
{
  currentState = origState;
  [self activateState];
  return self;
}

- setEcho
{
  currentState.sg_flags |= ECHO;
  [self activateState];
  return self;
}

- unSetEcho
{
  currentState.sg_flags &= (~ECHO);
  [self activateState];
  return self;
}

- setRaw
{
  currentState.sg_flags |= RAW;
  [self activateState];
  return self;
}

- unSetRaw
{
  currentState.sg_flags &= (~RAW);
  [self activateState];
  return self;
}

- setCbreak
{
  currentState.sg_flags |= CBREAK;
  [self activateState];
  return self;
}

- unSetCbreak
{
  currentState.sg_flags &= (~CBREAK);
  [self activateState];
  return self;
}

- setCrmod
{
  currentState.sg_flags |= CRMOD;
  [self activateState];
  return self;
}

- unSetCrmod
{
  currentState.sg_flags &= (~CRMOD);
  [self activateState];
  return self;
}

- activateState
{
  int cc;
  cc=ioctl(fd,TIOCSETP,&currentState);
  return self;
}


static struct setbps{
  int speed;
  int speedCode;
} speeds[] = {{0,B0},{50,B50},{75,B75},{110,B110},
		{134,B134},{150,B150},{200,B200},
		{300,B300},{600,B600},{1200,B1200},
		{1800,B1800},{2400,B2400},
		{4800,B4800},{9600,B9600}};
- setSpeed:(int)bps
{
  int i;
  for (i=0; i<sizeof(speeds)/sizeof(struct setbps); i++) {
    if (bps==speeds[i].speed) {
      currentState.sg_ispeed = speeds[i].speedCode;
      currentState.sg_ospeed = speeds[i].speedCode;
      return [self activateState];
    }
  }
  /* if get to here, requested speed was illegal */
  fprintf(stderr,"Tty method setSpeed: illegal speed %d\n",bps);
  return self;
}

- (int) getSpeed
{
  int i, sp;
  sp = currentState.sg_ospeed;
  for (i=0; i<sizeof(speeds)/sizeof(struct setbps); i++) {
    if (sp==speeds[i].speedCode)
      return speeds[i].speed;
  }
  fprintf(stderr,"Tty method getSpeed: currentState.sp_ospeed = %d unknown\n",sp);
  return 0;
}

- setParity:(int)parity
{
  par=parity;
  switch (parity) {
  case EVENP:
    currentState.sg_flags |= EVENP;
    currentState.sg_flags &= (~ODDP);
    break;
  case ODDP:
    currentState.sg_flags |= ODDP;
    currentState.sg_flags &= (~EVENP);
    break;
  case NOPARITY:
    currentState.sg_flags |= EVENP;
    currentState.sg_flags |= ODDP;
    break;
  default:
    fprintf(stderr,"Tty method setParity: unrecognized parity:%d\n",parity);
    fprintf(stderr,"should be EVENP=%d, ODDP=%d, or NOPARITY=%d\n",EVENP,ODDP,NOPARITY);
    par=NOPARITY;
  }
  return [self activateState];
}

- (int) getParity
{
  return par;
}

- (int) getfd
{
  return fd;
}

- (int) queuedChars
{
  int numchars;
  ioctl(fd,FIONREAD,&numchars);
  return numchars;
}

- (int) readInto:(char *)buf
{
  int i;
  i = read(fd,buf,[self queuedChars]);
  buf[i]=0;
  return (i);
}

- (int) writeOut:(char *)buf
{
  int i, len=strlen(buf);
  char *table;
  static char oddtab[128] = {
    0x80, 0x01, 0x02, 0x83, 0x04, 0x85, 0x86, 0x07, 
    0x08, 0x89, 0x8a, 0x0b, 0x8c, 0x0d, 0x0e, 0x8f, 
    0x10, 0x91, 0x92, 0x13, 0x94, 0x15, 0x16, 0x97, 
    0x98, 0x19, 0x1a, 0x9b, 0x1c, 0x9d, 0x9e, 0x1f, 
    0x20, 0xa1, 0xa2, 0x23, 0xa4, 0x25, 0x26, 0xa7, 
    0xa8, 0x29, 0x2a, 0xab, 0x2c, 0xad, 0xae, 0x2f, 
    0xb0, 0x31, 0x32, 0xb3, 0x34, 0xb5, 0xb6, 0x37, 
    0x38, 0xb9, 0xba, 0x3b, 0xbc, 0x3d, 0x3e, 0xbf, 
    0x40, 0xc1, 0xc2, 0x43, 0xc4, 0x45, 0x46, 0xc7, 
    0xc8, 0x49, 0x4a, 0xcb, 0x4c, 0xcd, 0xce, 0x4f, 
    0xd0, 0x51, 0x52, 0xd3, 0x54, 0xd5, 0xd6, 0x57, 
    0x58, 0xd9, 0xda, 0x5b, 0xdc, 0x5d, 0x5e, 0xdf, 
    0xe0, 0x61, 0x62, 0xe3, 0x64, 0xe5, 0xe6, 0x67, 
    0x68, 0xe9, 0xea, 0x6b, 0xec, 0x6d, 0x6e, 0xef, 
    0x70, 0xf1, 0xf2, 0x73, 0xf4, 0x75, 0x76, 0xf7, 
    0xf8, 0x79, 0x7a, 0xfb, 0x7c, 0xfd, 0xfe, 0x7f};
  static char eventab[128] = {
    0x00, 0x81, 0x82, 0x03, 0x84, 0x05, 0x06, 0x87, 
    0x88, 0x09, 0x0a, 0x8b, 0x0c, 0x8d, 0x8e, 0x0f, 
    0x90, 0x11, 0x12, 0x93, 0x14, 0x95, 0x96, 0x17, 
    0x18, 0x99, 0x9a, 0x1b, 0x9c, 0x1d, 0x1e, 0x9f, 
    0xa0, 0x21, 0x22, 0xa3, 0x24, 0xa5, 0xa6, 0x27, 
    0x28, 0xa9, 0xaa, 0x2b, 0xac, 0x2d, 0x2e, 0xaf, 
    0x30, 0xb1, 0xb2, 0x33, 0xb4, 0x35, 0x36, 0xb7, 
    0xb8, 0x39, 0x3a, 0xbb, 0x3c, 0xbd, 0xbe, 0x3f, 
    0xc0, 0x41, 0x42, 0xc3, 0x44, 0xc5, 0xc6, 0x47, 
    0x48, 0xc9, 0xca, 0x4b, 0xcc, 0x4d, 0x4e, 0xcf, 
    0x50, 0xd1, 0xd2, 0x53, 0xd4, 0x55, 0x56, 0xd7, 
    0xd8, 0x59, 0x5a, 0xdb, 0x5c, 0xdd, 0xde, 0x5f, 
    0x60, 0xe1, 0xe2, 0x63, 0xe4, 0x65, 0x66, 0xe7, 
    0xe8, 0x69, 0x6a, 0xeb, 0x6c, 0xed, 0xee, 0x6f, 
    0xf0, 0x71, 0x72, 0xf3, 0x74, 0xf5, 0xf6, 0x77, 
    0x78, 0xf9, 0xfa, 0x7b, 0xfc, 0x7d, 0x7e, 0xff};
  if (par==NOPARITY) {
    return write(fd,buf,len);
  }
  else
    table=(par==EVENP)? eventab : oddtab;
  for (i=0; i<len; i++) {
    if (write(fd, table+(*buf++), 1)<0)
      return -1;
  }
  return len;
}

@end
